/***************************************************************************
 *   Copyright (C) 2006 by SukkoPera                                       *
 *   sukkopera@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib/gstdio.h>
#include "datfile.h"
#include "romentry.h"
#include "file.h"
#include "filelist.h"
#include "zipfile.h"


/* PROGRAM OPTIONS */
typedef struct {
	gchar **dats;
	gboolean mode_identify;
	gboolean mode_rename;
	gchar *copy_dest;
	gchar *move_dest;
	gboolean zip_output;
	gboolean show_unknown;
	gboolean verbose;
} globals_s;


static GArray *read_dats (char **filename, guint32 *total_roms_no_out) {
	int i;
	datfile *rcf;
	GArray *out;
	guint32 total_roms_no;

	out = g_array_new (FALSE, FALSE, sizeof (datfile *));
	/* Calculate number of total known ROMs */
	total_roms_no = 0;
	g_fprintf (stderr, "--- Reading DAT files ---\n");
	for (i = 0; i < g_strv_length (filename); i++) {
		if (!(rcf = datfile_new_from_filename (filename[i]))) {
			g_fprintf (stderr, "Unable to read DAT file \"%s\"\n", filename[i]);
		} else {
			g_array_append_val (out, rcf);
			total_roms_no += rcf -> roms_no;
		}
		g_fprintf (stderr, "\n");
		
	}
	g_fprintf (stderr, "Read %u DAT file(s), for %u total known ROMs\n\n", out -> len, total_roms_no);

	/* Return total number of ROMs, if needed */
	if (total_roms_no_out)
		*total_roms_no_out = total_roms_no;

	return (out);
}


static GList *build_filelist (int dirc, char *dirv[]) {
	GList *files;
	int i;

	g_fprintf (stderr, "--- Building filelist ---\n");
	files = NULL;
	for (i = 0; i < dirc; i++) {
		g_fprintf (stderr, "Directory \"%s\"\n", dirv[i]);
		files = add_to_filelist (files, dirv[i], TRUE);
	}
	g_fprintf (stderr, "Total files to process: %d\n\n", g_list_length (files));

	return (files);
}


void go_identify (GArray *dats, guint32 total_roms_no, GList *files, gboolean show_unknown) {
	GList *cur, *unknown;
	filez *f;
	romentry *r;
	guint32 owned;
	datfile *rcf;
	int i;

	g_fprintf (stderr, "--- Processing filelist ---\n");
	owned = 0;
	unknown = NULL;
	for (cur = g_list_first (files); cur; cur = g_list_next (cur)) {
		f = (filez *) cur -> data;
		file_compute_sums (f);
// 		g_print ("Searching for %s (%s)\n", f -> filename, f -> crc32);
		for (i = 0; i < dats -> len; i++) {
			rcf = g_array_index (dats, datfile *, i);
			r = datfile_find (rcf, f);
			if (r) {
				g_printf ("%s -> %s\n", f -> filename, r -> filename);
				owned++;
				break;			/* First match is good! */
			} else {
				/* Append to unknown files list */
				unknown = g_list_append (unknown, f);
			}
		}
	}
	g_printf ("Correctly identified %u of %u known ROMs\n", owned, total_roms_no);

	if (show_unknown) {
		g_printf ("\n--- Unknown files ---\n");
		for (cur = g_list_first (unknown); cur; cur = g_list_next (cur)) {
			f = (filez *) cur -> data;
			g_printf ("%s [%s]\n", f -> filename, f -> crc32);
		}
		g_printf ("Found %u unknown files\n", g_list_length (unknown));
	}
	g_list_free (unknown);

	return;
}

typedef enum {
	CMOP_COPY,
	CMOP_MOVE,
} copymove_operation;


static gboolean file_copy_move_wrapper (filez *f, gchar *destfile, copymove_operation op, gboolean zip_output) {
	gboolean out;
	char *basename, *zipname, *zippandum;
	zipfile *zf;

	if (zip_output) {
		/* Make up ZIP filename */
		char *dot = strrchr (destfile, '.');
		g_assert (dot);
		basename = g_strndup (destfile, dot - destfile);
		zipname = g_strconcat (basename, ".zip", NULL);
		g_free (basename);

		if (g_file_test (zipname, G_FILE_TEST_EXISTS)) {
			/* We don't want to overwrite in any case */
			g_printf ("Destination file already exists, skipping\n");
			out = FALSE;
		} else if (f -> parent_archive) {
			/* File is part of an archive, we must extract it to a temp file and zip it again alone */
			basename = strstr (f -> filename, f -> parent_archive);
			if (basename) {
				basename += strlen (f -> parent_archive) + 1;		/* Uhm uhm... */

				/* Get a temp file name (a bit of a hack...) */
				zippandum = g_strdup ("rcctmp.XXXXXX");
				close (g_mkstemp (zippandum));
				g_unlink (zippandum);
//	 			g_debug ("Extracting \"%s\" from \"%s\" and saving as \"%s\"", basename, f -> parent_archive, zippandum);
				zf = zipfile_new (f -> parent_archive);
				zipfile_extract_one (zf, basename, zippandum, TRUE, FALSE);
				zf = zipfile_destroy (zf);
				out = TRUE;		// FIXME
			} else {
				g_assert_not_reached ();
			}
		} else {
			/* The file to be zipped is the original one */
			zippandum = g_strdup (f -> filename);
			out = TRUE;
		}

		if (out) {
			/* Go! */
			g_debug ("Zipping \"%s\" to \"%s\"", zippandum, zipname);
			zf = zipfile_new (zipname);
			if (zipfile_begin_add (zf)) {
				/* Make up the compressed file name into the ZIP (i. e. without path) */
				basename = g_path_get_basename (destfile);
				out = zipfile_add (zf, zippandum, basename, NULL);
				out &= zipfile_end_add (zf);
				g_free (basename);
			} else {
				/* Error opening ZIP file for writing */
				g_warning ("Couldn't open ZIP file \"%s\" for writing", zipname);
				out = FALSE;
			}
			zf = zipfile_destroy (zf);
			if (f -> parent_archive || op == CMOP_MOVE) {
				g_debug ("Removing \"%s\"", zippandum);
				g_unlink (zippandum);
			}
			g_free (zippandum);
		}
		g_free (zipname);
	} else {
		/* Output is not compressed */
		if (g_file_test (destfile, G_FILE_TEST_EXISTS)) {
			/* We don't want to overwrite in any case */
			g_printf ("Destination file already exists, skipping\n");
			out = FALSE;
		} else if (f -> parent_archive) {
			/* File is part of an archive, we must extract it */
			basename = strstr (f -> filename, f -> parent_archive);
			if (basename) {
				basename += strlen (f -> parent_archive) + 1;		/* Uhm uhm... */
	// 			g_debug ("Extracting \"%s\" from \"%s\" and saving as \"%s\"", basename, f -> parent_archive, dest);
				zf = zipfile_new (f -> parent_archive);
				zipfile_extract_one (zf, basename, destfile, TRUE, FALSE);
				zf = zipfile_destroy (zf);
				out = TRUE;		// FIXME
			} else {
				g_assert_not_reached ();
			}
		} else {
			if (op == CMOP_COPY)
				out = filecopy (f -> filename, destfile);
			else if (op == CMOP_MOVE)
				out = filemove (f -> filename, destfile);
			else
				g_assert_not_reached ();
		}
	}

	return (out);
}


void go_copy_move (GArray *dats, guint32 total_roms_no, GList *files, gchar *destdir, copymove_operation op, gboolean zip_output) {
	GList *cur, *unknown;
	filez *f;
	romentry *r;
	guint32 owned;
	gchar *dest;
	datfile *rcf;
	int i;

	if (op == CMOP_COPY)
		g_fprintf (stderr, "--- Copy mode selected ---\n");
	else if (op == CMOP_MOVE)
		g_fprintf (stderr, "--- Move mode selected ---\n");
	else
		g_assert_not_reached ();

	/* Setup destination directory tree */
	if (g_file_test (destdir, G_FILE_TEST_IS_DIR)) {
		g_fprintf (stderr, "Destination directory \"%s\" already exists\n", destdir);
	} else {
		g_fprintf (stderr, "Creating destination directory \"%s\"\n", destdir);
		g_mkdir (destdir, 0755);
	}
	if (dats -> len > 1) {
		/* If we are using multiple DATs, make one more subdirectory level, to distinguish
		   which DAT the ROM was found into. */
		for (i = 0; i < dats -> len; i++) {
			rcf = g_array_index (dats, datfile *, i);
			dest = g_build_filename (destdir, rcf -> emulator_version, NULL);	// FIXME If emulator_version contains weird characters...
			if (g_file_test (dest, G_FILE_TEST_IS_DIR)) {
				g_fprintf (stderr, "Destination subdirectory \"%s\" already exists\n", dest);
			} else {
				g_fprintf (stderr, "Creating destination subdirectory \"%s\"\n", dest);
				g_mkdir (dest, 0755);
			}
			g_free (dest);
		}
	}
	g_fprintf (stderr, "\n");

	/* Process files */
	g_fprintf (stderr, "--- Processing filelist ---\n");
	owned = 0;
	unknown = NULL;
	for (cur = g_list_first (files); cur; cur = g_list_next (cur)) {
		f = (filez *) cur -> data;
		file_compute_sums (f);		// FIXME See if already computed (i. e. ZIP files)
// 		g_print ("Searching for %s (%s)\n", f -> filename, f -> crc32);
		for (i = 0; i < dats -> len; i++) {
			rcf = g_array_index (dats, datfile *, i);
			r = datfile_find (rcf, f);
			if (r) {
				(rcf -> found_roms)++;
				g_printf ("%s -> %s\n", f -> filename, r -> filename);
				if (dats -> len > 1) {
					/* If we are using multiple DATs... */
					dest = g_build_filename (destdir, rcf -> emulator_version, r -> filename, NULL);
				} else {
					dest = g_build_filename (destdir, r -> filename, NULL);
				}
				if (!file_copy_move_wrapper (f, dest, op, zip_output)) {
					g_fprintf (stderr, "Cannot write to \"%s\"\n", dest);
				} else {
					owned++;
					break;
				}
				g_free (dest);
			} else {
				/* Append to unknown files list */
				unknown = g_list_append (unknown, f);
			}
		}
	}
	g_printf ("Correctly copied/moved %u of %u known ROMs\n", owned, total_roms_no);
	for (i = 0; i < dats -> len; i++) {
		rcf = g_array_index (dats, datfile *, i);
		g_printf ("%s: %u/%u\n", rcf -> filename, rcf -> found_roms, rcf -> roms_no);
	}
	g_printf ("\n");

	return;
}


static gboolean cmdline_parse (globals_s *globals, int *argc, char **argv[]) {
	gboolean out;
	GError *error;
	GOptionContext *context;
	GOptionEntry entries[] = {
		{"datfile", 'd', 0, G_OPTION_ARG_FILENAME_ARRAY, &(globals -> dats), "Load DAT file (can be used multiple times)", NULL},
		{"identify", 'i', 0, G_OPTION_ARG_NONE, &(globals -> mode_identify), "Operation mode: Identify ROMs, don't rename or move", NULL},
// 		{"rename", 'r', 0, G_OPTION_ARG_NONE, &(globals -> mode_rename), "Operation mode: Identify ROMs and rename them inplace", NULL},
		{"copy", 'c', 0, G_OPTION_ARG_FILENAME, &(globals -> copy_dest), "Operation mode: Identify ROMs and copy them to the specified directory", "directory"},
		{"move", 'm', 0, G_OPTION_ARG_FILENAME, &(globals -> move_dest), "Operation mode: Identify ROMs and move them to the specified directory", "directory"},
		{"zipout", 'z', 0, G_OPTION_ARG_NONE, &(globals -> zip_output), "ZIP-compress output files", NULL},
		{"unknown", 'u', 0, G_OPTION_ARG_NONE, &(globals -> show_unknown), "Show unknown files", NULL},
		{"verbose", 'v', 0, G_OPTION_ARG_NONE, &(globals -> verbose), "Be verbose", NULL},
		{NULL}
	};

	memset (globals, 0, sizeof (globals_s));
	context = g_option_context_new ("- Rename ROMs according to RomCenter DAT files");
	g_option_context_add_main_entries (context, entries, NULL);
	g_option_context_set_help_enabled (context, TRUE);
	g_option_context_set_ignore_unknown_options (context, FALSE);
	error = NULL;
	g_option_context_parse (context, argc, argv, &error);

	/* Sanity checks */
	if (globals -> mode_identify + globals -> mode_rename + (globals -> copy_dest != NULL) + (globals -> move_dest != NULL) != 1) {
		g_fprintf (stderr, "ERROR: Please specify exactly one operation mode\n");
		out = FALSE;
	} else if (!(globals -> dats)) {
		g_fprintf (stderr, "ERROR: No DAT file specified\n");
		out = FALSE;
	} else {
		out = TRUE;
	}
	
	return (out);
}


int main (int argc, char *argv[]) {
	globals_s globals;
	int out;
	GList *files;
	GTimer *timer;
	GArray *dat_files;
	guint32 total_roms_no;

	/* Welcome text */
	g_fprintf (stderr, "ROM Collection Checker (" PACKAGE ") " VERSION "\n"
			"Copyright (C) 2006 SukkoPera <sukkopera@gmail.com>\n"
			"This software comes with ABSOLUTELY NO WARRANTY.\n"
			"This is free software, and you are welcome to redistribute it\n"
			"under certain conditions; see COPYING for details.\n\n");
	fflush (stderr);
	
	if ((out = cmdline_parse (&globals, &argc, &argv))) {
		dat_files = read_dats (globals.dats, &total_roms_no);
		files = build_filelist (argc - 1, argv + 1);
		timer = g_timer_new ();
		if (globals.mode_identify) {
			go_identify (dat_files, total_roms_no, files, globals.show_unknown);
			out = EXIT_SUCCESS;
		} else if (globals.copy_dest) {
			go_copy_move (dat_files, total_roms_no, files, globals.copy_dest, CMOP_COPY, globals.zip_output);
			out = EXIT_SUCCESS;
		} else if (globals.move_dest) {
			go_copy_move (dat_files, total_roms_no, files, globals.move_dest, CMOP_MOVE, globals.zip_output);
			out = EXIT_SUCCESS;
		} else {
			g_fprintf (stderr, "ERROR: Operation mode not (yet) supported\n");
			out = EXIT_FAILURE;
		}

		if (out == EXIT_SUCCESS) {
			g_timer_stop (timer);
			g_fprintf (stderr, "\nTime elapsed: %lf seconds\n", g_timer_elapsed (timer, NULL));
		}
		g_timer_destroy (timer);
	}
	
	return (out);
}
